----------------------------------------------------------------------------
-- Copyright (c) 2015 Turishev Evgeniy <ew_turi@mail.ru>                  --
-- This file is licensed under the GNU General Public License version 3   --
--  or any later version.                                                 --
----------------------------------------------------------------------------

with Interfaces; use Interfaces;
with Oled.Draw;

package body Oled.Buffered is

   function Min (X1, X2 : Unsigned_8) return Unsigned_8
   renames Unsigned_8'Min;

   function Max (X1, X2 : Unsigned_8) return Unsigned_8
   renames Unsigned_8'Max;

   procedure Row_And_Bit (Y        : in     Unsigned_8;
                          Row, Bit :    out Unsigned_8);
   pragma Inline (Row_And_Bit);

   procedure Set_Byte (Bmp   : in out Bitmap;
                       Row   : in     Unsigned_8;
                       X     : in     Unsigned_8;
                       Byte  : in     Unsigned_8;
                       Mask  : in     Unsigned_8;
                       Oper  : in     Bit_Oper := Bit_Copy);


   -----------------
   -- Bitmap_Copy --
   -----------------

   procedure Bitmap_Copy (Dst_Bmp : in out Bitmap;
                          Src_Bmp : in     Bitmap;
                          Dst_X   : in     Unsigned_8;
                          Dst_Y   : in     Unsigned_8;
                          Oper    : in     Bit_Oper := Bit_Copy)
   is
   begin
      Bitmap_Copy (Dst_Bmp, Src_Bmp,
                   Dst_X => Dst_X,
                   Dst_Y => Dst_Y,
                   Src_X1 => Src_Bmp'First (2),
                   Src_Y1 => Src_Bmp'First (1),
                   Src_X2 => Src_Bmp'Last (2),
                   Src_Y2 => (Src_Bmp'Last (1) + 1) * 8 - 1,
                   Oper => Oper);
   end Bitmap_Copy;

   procedure Bitmap_Copy (Dst_Bmp                        : in out Bitmap;
                          Src_Bmp                        : in     Bitmap;
                          Dst_X, Dst_Y                   : in Unsigned_8;
                          Src_X1, Src_Y1, Src_X2, Src_Y2 : in Unsigned_8;
                          Oper : in Bit_Oper := Bit_Copy)
   is
      Src_X_First : constant Unsigned_8 := Min (Src_X1, Src_X2);
      Src_X_Last  : constant Unsigned_8 := Max (Src_X1, Src_X2);
      Dst_X_Last  : constant Unsigned_8 := Dst_X + Src_X_Last - Src_X_First;
      Src_Row_First, Src_Row_Last, Src_Bit_First, Src_Bit_Last  : Unsigned_8;
      Dst_Row_First, Dst_Row_Last, Dst_Bit_First, Dst_Bit_Last  : Unsigned_8;

   begin
      Row_And_Bit (Min (Src_Y1, Src_Y2), Src_Row_First, Src_Bit_First);
      Row_And_Bit (Max (Src_Y1, Src_Y2), Src_Row_Last, Src_Bit_Last);

      if not (Src_Row_First in Src_Bmp'Range)
        or else not (Src_Row_Last in Src_Bmp'Range)
        or else not (Src_X1 in Src_Bmp'Range (2))
        or else not (Src_X2 in Src_Bmp'Range (2))
      then
         raise Program_Error with "source parameters error";
      end if;

      Row_And_Bit (Dst_Y, Dst_Row_First, Dst_Bit_First);
      Row_And_Bit (Dst_Y + Unsigned_8 (abs (Integer (Src_Y2)
                                              - Integer (Src_Y1))),
                   Dst_Row_Last, Dst_Bit_Last);

      if not (Dst_Row_First in Dst_Bmp'Range)
        or else not (Dst_Row_Last in Dst_Bmp'Range)
        or else not (Dst_X in Dst_Bmp'Range (2))
        or else not (Dst_X_Last in Dst_Bmp'Range (2))
      then
         raise Program_Error with "destination parameters error";
      end if;

      declare
         Bit_Shift : constant Integer :=
           (Integer (Dst_Bit_First) - Integer (Src_Bit_First));
         Mask : Unsigned_8;
         U16  : Unsigned_16;
         Src_Row : Unsigned_8 := Src_Row_First;
         Src_X : Unsigned_8;
         Src_Byte : Unsigned_8;
      begin
         for R in Dst_Row_First .. Dst_Row_Last loop

            if R = Dst_Row_First then
               Mask := Shift_Left (16#FF#, Integer (Dst_Bit_First));

            elsif R = Dst_Row_Last then
               declare
                  M : Unsigned_8 :=
                    Shift_Right (16#FF#, 7 - Integer (Dst_Bit_Last));
               begin
                  if Dst_Row_Last = Dst_Row_First then -- only one row
                     Mask := Mask and M;
                  else
                     Mask := M;
                  end if;
               end;
            else
               Mask := 255;
            end if;

            Src_X := Src_X_First;

            for X in Dst_X .. Dst_X_Last loop
               U16 := 0;

               if Bit_Shift = 0 then -- simple copy row to row
                  Src_Byte := Src_Bmp (Src_Row, Src_X);

               elsif Bit_Shift > 0 then
                  if Src_Row = Src_Row_First then
                     Src_Byte := Shift_Left (Src_Bmp (Src_Row, Src_X),
                                             Bit_Shift);

                  elsif Src_Row > Src_Row_Last then
                     Src_Byte := Shift_Right (Src_Bmp (Src_Row - 1, Src_X),
                                              8 - Bit_Shift);

                  else -- Src_Row_Last >=  Src_Row > Src_Row_First
                       -- need use 2 src-byte with shift
                     U16 := Shift_Left (Unsigned_16
                                          (Src_Bmp (Src_Row, Src_X)), 8)
                       or Unsigned_16 (Src_Bmp (Src_Row - 1, Src_X));

                     Src_Byte := Unsigned_8 (Shift_Right (U16, 8 - Bit_Shift)
                                               and 16#00FF#);
                  end if;

               else  --  Bit_Shift < 0
                  if Src_Row = Src_Row_Last then -- use only last src-byte
                     Src_Byte := Shift_Right (Src_Bmp (Src_Row, Src_X),
                                              -Bit_Shift);

                  else --  Src_Row < Src_Row_Last
                       --  use 2 src-byte with shift
                     U16 := Shift_Left (Unsigned_16
                                          (Src_Bmp (Src_Row + 1, Src_X)), 8)
                       or Unsigned_16 (Src_Bmp (Src_Row, Src_X));

                     Src_Byte := Unsigned_8 (Shift_Right (U16, -Bit_Shift)
                                               and 16#00FF#);
                  end if;
               end if;

               Set_Byte (Bmp => Dst_Bmp, Row => R, X => X,
                         Byte => Src_Byte, Mask => Mask, Oper => Oper);

               Src_X := Src_X + 1;
            end loop;
            Src_Row := Src_Row + 1;
         end loop;
      end;

   end Bitmap_Copy;

   ------------------
   -- Clear_Bitmap --
   ------------------

   procedure Clear_Bitmap (Bmp  : in out Bitmap;
                           Byte : in     Unsigned_8 := 0) is
   begin
      for R in Bmp'First .. Bmp'Last loop
         for C in Bmp'First (2) .. Bmp'Last (2) loop
            Bmp (R, C) := Byte;
         end loop;
      end loop;
   end Clear_Bitmap;

   ---------------------
   -- Draw_Horiz_Line --
   ---------------------

   procedure Draw_Horiz_Line (Bmp        : in out Bitmap;
                              Y, X1, X2  : in     Unsigned_8;
                              Oper       : in Bit_Oper := Bit_Copy)
   is
      Row, Bit, B : Unsigned_8;
   begin
      Row_And_Bit (Y, Row, Bit);
      B := Shift_Left (1, Integer (Bit));

      for X in Min (X1, X2) .. Max (X1, X2) loop
         Set_Byte (Bmp => Bmp, Row => Row, X => X,
                   Byte => B, Mask => B, Oper => Oper);
      end loop;
   end Draw_Horiz_Line;

  ---------------
   -- Draw_Rect --
   ---------------

   procedure Draw_Rect (Bmp             : in out Bitmap;
                        X1, Y1, X2, Y2  : in     Unsigned_8;
                        Oper            : in Bit_Oper := Bit_Copy)
   is
      Row_1, Row_2, Bit_1, Bit_2, Mask_1, Mask_2 : Unsigned_8;
   begin
      Row_And_Bit (Min (Y1, Y2), Row_1, Bit_1);
      Row_And_Bit (Max (Y1, Y2), Row_2, Bit_2);

      -- mask for first row
      Mask_1 := Shift_Left (16#FF#, Integer (Bit_1));
      -- mask for last row
      Mask_2 := Shift_Right (16#FF#, 7 - Integer (Bit_2));

      if Row_2 = Row_1 then
         Mask_2 := Mask_2 and Mask_1;
      else
         for X in Min (X1, X2) .. Max (X1, X2) loop
            Set_Byte (Bmp => Bmp, Row => Row_1, X => X,
                      Byte => Mask_1, Mask => Mask_1, Oper => Oper);
         end loop;
      end if;

      for X in Min (X1, X2) .. Max (X1, X2) loop
         Set_Byte (Bmp => Bmp, Row => Row_2, X => X,
                   Byte => Mask_2, Mask => Mask_2, Oper => Oper);
      end loop;

      if Row_2 - Row_1 > 1 then
         for J in Row_1 + 1 .. Row_2 - 1 loop
            for X in Min (X1, X2) .. Max (X1, X2) loop
               Set_Byte (Bmp => Bmp, Row => J, X => X,
                         Byte => 255, Mask => 255, Oper => Oper);
            end loop;
         end loop;
      end if;
   end Draw_Rect;

   ------------------
   -- Draw_Vertic_Line --
   ------------------

   procedure Draw_Vertic_Line (Bmp       : in out Bitmap;
                               X, Y1, Y2 : in     Unsigned_8;
                               Oper      : in Bit_Oper := Bit_Copy) is
   begin
      Draw_Rect (Bmp, X1 => X, Y1 => Y1, X2 => X, Y2 => Y2, Oper => Oper);
   end Draw_Vertic_Line;

   -----------------
   -- Row_And_Bit --
   -----------------

   procedure Row_And_Bit (Y        : in     Unsigned_8;
                          Row, Bit :    out Unsigned_8) is
   begin
      Row  := Y / 8;
      Bit  := Unsigned_8 (Integer (Y) mod 8);
   end Row_And_Bit;

   --------------
   -- Set_Byte --
   --------------

   procedure Set_Byte (Bmp   : in out Bitmap;
                       Row   : in     Unsigned_8;
                       X     : in     Unsigned_8;
                       Byte  : in     Unsigned_8;
                       Mask  : in     Unsigned_8;
                       Oper  : in     Bit_Oper := Bit_Copy)
   is
      DB : constant Unsigned_8 := Bmp (Row, X);
   begin
      if Oper = Bit_Copy then
         Bmp (Row, X) := (DB and not Mask)
           or (Byte and Mask);
      elsif Oper = Bit_Copy_Inverse then
         Bmp (Row, X) := (DB and not Mask)
           or ((not Byte) and Mask);
      elsif Oper = Bit_Xor then
         Bmp (Row, X) := DB xor (Byte and Mask);
      elsif Oper = Bit_Or then
         Bmp (Row, X) := DB or (Byte and Mask);
      elsif Oper = Bit_And then
         Bmp (Row, X) := DB and (Byte or not Mask);
      else
         raise Program_Error with "operation not realized";
      end if;
   end Set_Byte;

   ---------------
   -- Set_Pixel --
   ---------------

   procedure Set_Pixel (Bmp   : in out Bitmap;
                        X, Y  : in     Unsigned_8;
                        Oper  : in     Bit_Oper := Bit_Copy)
   is
      Row, Bit, V    : Unsigned_8;
   begin
      Row_And_Bit (Y, Row, Bit);

      V := Shift_Left (1, Integer (Bit));

      Set_Byte (Bmp => Bmp, Row => Row, X => X,
                Byte => V, Mask => V, Oper => Oper);
   end Set_Pixel;

end Oled.Buffered;
