----------------------------------------------------------------------------
-- Copyright (c) 2015 Turishev Evgeniy <ew_turi@mail.ru>                  --
-- This file is licensed under the GNU General Public License version 3   --
--  or any later version.                                                 --
----------------------------------------------------------------------------

with SAM4S; use SAM4S;
with System; use System;
with Interfaces; use Interfaces;
with System.Machine_Code; use System.Machine_Code;
with System.Storage_Elements;
with Timer;

package body SAM4S_Init is

   Main_Thread_Prior : constant Priority := 1;

   NL : constant String := ASCII.LF & ASCII.HT;

   procedure Fault_Handler_Wrapper;
   pragma Export (Asm, Fault_Handler_Wrapper, "Fault_Handler");
   procedure IRQ_Handler_Wrapper;
   pragma Export (Asm, IRQ_Handler_Wrapper, "IRQ_Handler");

   procedure Set_Traps_Prior;

   procedure Default_IRQ_Handler (IRQ : IRQ_Id);
   function Pri_To_Byte (P : Priority) return Byte;


   IRQ_Handlers_Table : array (IRQ_Id'Range) of IRQ_Handler
     := (others => Default_IRQ_Handler'Access);

   -------------------------
   -- Default_Irq_Handler --
   -------------------------

   procedure Default_IRQ_Handler (IRQ : IRQ_Id) is
   begin
      null;
   end Default_IRQ_Handler;

   -----------------
   -- Pri_To_Byte --
   -----------------

   function Pri_To_Byte (P : Priority) return Byte is
   begin
      return Byte (Interrupt_Priority'Last - P + 1) * 16;
   end Pri_To_Byte;

   ----------
   -- Init --
   ----------

   procedure Init (Main_Priority : Priority := 1) is
   begin
      Disable_Interrupts;
       --  Mask Fault
      Asm ("cpsid f", Volatile => True);

      Set_Traps_Prior;

      Timer.Init;
       --  Unmask Fault
      Asm ("cpsie f", Volatile => True);

      Set_Current_Priority (Main_Priority);

      Enable_Interrupts;
   end Init;

   procedure Fault_Handler_Wrapper is
   begin
      null;
   end Fault_Handler_Wrapper;

   ---------------------
   -- Set_Traps_Prior --
   ---------------------

   procedure Set_Traps_Prior is
   begin
      SYS_CTRL.CCR := SYS_CTRL.CCR or 16#211#;

      --  Set priorities of system handlers.
      --  All other system handlers run at the highest
      --  priority (0), so they will not be interrupted.

      SYS_CTRL.SHPR1 := (0, 0, 0, 0);
      SYS_CTRL.SHPR2 := (0, 0, 0, 0);
      SYS_CTRL.SHPR3 := (0, 0, 0, 0);
      --  SYS_CTRL.SHPR3 (3) - SysTick - in Timer module

      --  Write the required key (16#05FA#) and desired PRIGROUP value. We
      --  configure this to 3, to have 16 group priorities

      SYS_CTRL.AIRCR := 16#05FA_0300#;
      pragma Assert (SYS_CTRL.AIRCR = 16#FA05_0300#); --  Key value is swapped

      --  enable bus and memory fault
      SYS_CTRL.SHCSR := SYS_CTRL.SHCSR or 16#7_0000#;
   end Set_Traps_Prior;

   -------------------------
   -- IRQ_Handler_Wrapper --
   -------------------------

   procedure IRQ_Handler_Wrapper
   is
      W   : Word;
      IRQ : IRQ_Id;
   begin
      --  get IRQ_Id
      Asm ("mrs %0, ipsr",
           Word'Asm_Output ("=r", W),
           Volatile => True);

      IRQ :=  Natural (W and 16#FF#) - 16;
      IRQ_Handlers_Table (IRQ).all (IRQ);
   end IRQ_Handler_Wrapper;

   ------------------------
   -- Disable_Interrupts --
   ------------------------

   procedure Disable_Interrupts is
   begin
      Asm ("cpsid i", Volatile => True);
   end Disable_Interrupts;

   -----------------------
   -- Enable_Interrupts --
   -----------------------

   procedure Enable_Interrupts is
   begin
      --  The following enables interrupts and will cause any pending
      --  interrupts to take effect. The barriers and their placing are
      --  essential, otherwise a blocking operation might not cause an
      --  immediate context switch, violating mutual exclusion.

      Asm ("cpsie i" & NL
         & "dsb"     & NL
         & "isb",
           Clobber => "memory", Volatile => True);
   end Enable_Interrupts;

   --------------------------
   -- Set_Current_Priority --
   --------------------------

   procedure Set_Current_Priority (Prior : Priority) is
   begin
      Asm ("msr BASEPRI, %0",
           Inputs => Byte'Asm_Input ("r", Pri_To_Byte (Prior)),
           Volatile => True);
   end Set_Current_Priority;

   ---------------------
   -- Set_IRQ_Handler --
   ---------------------

   procedure Set_IRQ_Handler (Id      : IRQ_Id;
                              Handler : IRQ_Handler;
                              Prior   : Priority := 8)
   is
      Reg_Num : constant Natural := Id / 32;
      Bit     : constant Word := 2** (Id mod 32);
   begin
      Disable_Interrupts;

      IRQ_Handlers_Table (Id) := Handler;

      -- set IRQ priority
      NVIC_IPR (Id) := Byte (Pri_To_Byte (Prior));
      -- enable IRQ in NVIC
      NVIC_ISER (Reg_Num) := Word (Bit);

      Enable_Interrupts;
   end Set_IRQ_Handler;

end SAM4S_Init;
